# -*- coding: utf-8 -*-

"""Console script for pyyacp."""

from pyyacp.datatable import parseDataTables

from pyyacp.profiler.profiling import apply_profilers

import click

from pyyacp.web.to_html import to_html_string

from pyjuhelpers.logging import *
from pyjuhelpers.timer import Timer


class NotRequiredIf(click.Option):
    def __init__(self, *args, **kwargs):
        self.not_required_if = kwargs.pop('not_required_if')
        assert self.not_required_if, "'not_required_if' parameter required"
        kwargs['help'] = (kwargs.get('help', '') +
            ' NOTE: This argument is mutually exclusive with %s' %
            self.not_required_if
        ).strip()
        super(NotRequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        we_are_present = self.name in opts
        other_present = self.not_required_if in opts

        if other_present:
            if we_are_present:
                raise click.UsageError(
                    "Illegal usage: `%s` is mutually exclusive with `%s`" % (
                        self.name, self.not_required_if))
            else:
                self.prompt = None

        return super(NotRequiredIf, self).handle_parse_result(
            ctx, opts, args)

import logging
logging.basicConfig(level=logging.INFO)
@click.group()
@click.option('-v', '--verbose', count=True, help="-v INFO, -vv DEBUG")
def cli(verbose):

    if verbose == 0:
        click.echo("Default logging")
        logging.config.dictConfig(defaultConf)
    elif verbose == 1:
        logging.config.dictConfig(infoConf)
        click.echo("Info logging")
    elif verbose == 2:
        click.echo("Debug logging")
        logging.config.dictConfig(debugConf)
    else:
        click.echo("Default & File logging")
        logging.config.dictConfig(fileConf)

    pass

@cli.command()
@click.argument("csv")
@click.option('-b', '--bench',  is_flag=True)
def inspect(csv, bench):
    """Inspect a CSV file to figure out about the dialect, comment and header lines and the overall structure."""

    tables = parseDataTables(csv)

    click.echo("Found {} tables".format(len(tables)))
    for i,table in enumerate(tables):
        click.echo("Table-{}".format(i))
        table.print_summary()

    if bench:
        click.echo("TIMING")
        click.echo(Timer.printStats())


@cli.command()
@click.argument("csv")
@click.option('-b', '--bench',  is_flag=True)
@click.option('--bench_out', help="file to dump benchmark", type=click.Path())
def clean(csv, bench, bench_out):
    """Parse and clean a CSV file (strip comments, utf-8 encoding, default dialect"""


    tables = parseDataTables(csv)

    if len(tables)>1:
        click.echo("Found {} tables".format(len(tables)))
    else:
        print(tables[0].generate())

    if bench:
        click.echo("TIMING")
        click.echo(Timer.printStats())
    if bench_out:
        Timer.to_csv(bench_out)

@cli.command()
@click.argument("csv")
@click.option('--html',  help='html representation',type=click.Path(resolve_path=True))
@click.option('-l','--load',  help='try to open/load html file',count=True)
@click.option('-s','--sample',  help='sample rows',type=int, default=5)
@click.option('-b', '--bench',  is_flag=True)
@click.option('--bench_out', help="file to dump benchmark", type=click.Path())
def profile(csv, html, load, sample, bench, bench_out):
    """Console script for pyyacp."""

    tables = parseDataTables(csv)

    click.echo("Found {} tables".format(len(tables)))
    for table in tables:
        ptable = apply_profilers(table)

        if html:
            click.echo("Writing HTML representation to {}".format(click.format_filename(html)))
            with open(html, "w") as f:
                f.write(to_html_string(ptable,sample=sample))
            if load:
                click.launch(html)
        else:
            ptable.print_summary()

    if bench:
        click.echo("TIMING")
        click.echo(Timer.printStats())
    if bench_out:
        Timer.to_csv(bench_out)

if __name__ == "__main__":
    cli() # pragma: no cover




# -*- coding: utf-8 -*-
from csvmimesis.table_generator import create_table, write_table, create_table_pair



table = {
        "seed":"1232",
        "prefix": None,
        "local": 'de',
        "rows": 1000,
        "columns": ['address.address', 'address.city']
    }

table1 = {
        "seed":"1231",
        "prefix": None,
        "local": 'de',
        "rows": 20,
        "columns": ['address.address', ['address.city',0.5], ["address.longitude",.5],["address.latitude",.5],["address.postal_code",.5]]
    }

def create_dummy_table(csv, table=table1):


    _tab = create_table(table)
    write_table(_tab, file=csv)


table_pair={
        "local":"de",
        "seed": "1232",
        "shared_providers":None,
        "add_providers":[ [ ["address.postal_code",1],["address.street_name",1], ["address.street_number",1]],
                          [ ["address.latitude",1], ["address.longitude",1],["address.country",1]]
                        ],
        "join_providers":[["address.city",1]],
        "rows":[20,20]
    }

table_pair1={
        "local":"en",
        "seed": "1232",
        "shared_providers":None,
        "add_providers":[ [ ["person.last_name",1],["person.name",1] ],
                          [ ["person.username",1], ["person.email",1],["address.country",1] ]
                        ],
        "join_providers":[["person.identifier",1]],
        "rows":[20,20],

    }

table_pair2={
        "local":"en",
        'seed':"2323",
        "shared_providers":[ ["address.city",1],["address.country",1],["address.postal_code",1]],
        "add_providers":None,
        "join_providers":None,
        "rows":[20,20]}


def create_dummy_table_pair(csv1, csv2, table_pair=table_pair):


    _tables = create_table_pair(**table_pair)
    write_table(_tables[0], file=csv1)
    write_table(_tables[1], file=csv2)






import importlib

from pyjuhelpers.basicconfig import Config

import pkg_resources
config_file = pkg_resources.resource_filename(__name__,"config.yaml")





from anycsv.config import _config as _anycsvconfig
_config=_anycsvconfig

_config = Config(config_file=config_file)
_config.merge(init=_config)

pyyacpconfig = _config.pyyacp
anycsvconfig = _config.anycsv




#overwrite anycsvconfig
#import anycsv
#anycsv.config.anycsvconfig = _config.anycsv




from jinja2 import Environment, FileSystemLoader
import os

PATH = os.path.dirname(os.path.abspath(__file__))
env = Environment(
    autoescape=False,
    loader=FileSystemLoader(os.path.join(PATH, 'templates'))
)

def to_html_string(table, sample=5):


    if sample:
        data = [list(table.data.columns)] + [list(row) for idx, row in table.data.head(sample).iterrows()]
    else:
        data = [list(table.data.columns)] + [list(row) for idx, row in table.data.iterrows()]

    _table = {
        'profiledata':table.profiledata(),
        'table':table,
        'sample':data
    }

    template = env.get_template('profile_view.jinja')
    return template.render(table=_table)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import operator
from collections import defaultdict

import numpy as np

from pyyacp.profiler import ColumnProfiler
from pyjuhelpers.timer import timer

__author__ = 'nina'

ColumnStatsProfilerResult = collections.namedtuple('ColumnStatsProfilerResult',
                                                  ['num_rows',
                                                   'min_value','max_value',
                                                    'min_len','max_len','mean_len','median_len',
                                                   'empty','distinct','uniqueness','mode','constancy'
                                                   ])
ColumnStatsProfilerResult.__new__.__defaults__ = (0,) * len(ColumnStatsProfilerResult._fields)
class ColumnStatsProfiler(ColumnProfiler):

    def __init__(self):
        super(ColumnStatsProfiler, self).__init__('csp', 'stats')
        self.vlen = []
        self.dv = defaultdict(int)

    @timer(key="profile.col_stats")
    def _profile_column(self, values, meta)->ColumnStatsProfilerResult:
        """

        :param values:
        :param meta:
        :return: instance of ColumnStatsProfilerResult
        """
        _vlena= self.vlen.append
        for v in values:
            _v = v.strip()
            _vlena(len(_v))
            self.dv[_v]+=1
        return self._compile_stats()

    def result_datatype(self):
        return ColumnStatsProfilerResult()

    def _compile_stats(self):
        stats={}
        stats['num_rows'] = len(self.vlen)

        a = np.array(self.vlen)
        an = a[a>0]

        if len(a[a==0])>0:
            if '' in self.dv:
                self.dv.pop('')

        stats['min_value'] = min(self.dv) if len(self.dv)>0 else ''
        stats['max_value'] = max(self.dv) if len(self.dv)>0 else ''


        stats['min_len'] = min(an) if len(an)>0 else 0
        stats['max_len'] = max(an) if len(an)>0 else 0
        stats['mean_len'] = np.mean(an)
        stats['median_len'] = np.median(an)

        stats['empty']= len(a[a==0])
        stats['distinct']=len(self.dv)
        stats['uniqueness']=len(self.dv)/float(len(a)) if len(a)>0 else 0

        sorted_values = sorted(self.dv.items(), key=operator.itemgetter(1), reverse=True)
        top_values = [(sorted_values[i][0], int(sorted_values[i][1])) for i in range(min(5, len(self.dv)))]

        stats['mode'] = sorted_values[0][0] if len(sorted_values)>0 else None

        stats['constancy'] = max(self.dv.values()) / float(len(a)) if len(self.dv)>0 and  len(a)>0 else 0

        return ColumnStatsProfilerResult(**stats)



from dtpattern.dtpattern1 import pattern
from pyyacp.profiler import ColumnProfiler
from pyjuhelpers.timer import timer

class ColumnPatternProfiler(ColumnProfiler):

    def __init__(self):
        super(ColumnPatternProfiler, self).__init__('cpp','pattern')

    def result_datatype(self):
        return str

    @timer(key="profile.col_pattern")
    def _profile_column(self, values, meta) -> str:
        res= pattern(values, includeCount=False)
        if len(res)>0:
            return res[0]






from pyyacp.datatable import create_column_names


def test_basic():

    rows= [['A','B','C']]
    names = create_column_names(rows)

    h_transposed = list(map(list, zip(*rows)))
    for i, hs in enumerate(h_transposed):
        if 'miss' in names[i]:
            assert all(h is None or len(h.strip()) > 0 for h in hs)
        else:
            assert names[i] in hs

def test_missing():

    rows= [['A',None,'C']]
    names = create_column_names(rows)

    h_transposed = list(map(list, zip(*rows)))
    for i, hs in enumerate(h_transposed):
        if 'miss' in names[i]:
            assert all(h is None or len(h.strip()) > 0 for h in hs)
        else:
            assert names[i] in hs


def test_multi():

    rows= [['A',None,'C'],['a','b','c']]
    names = create_column_names(rows)

    h_transposed = list(map(list, zip(*rows)))
    for i, hs in enumerate(h_transposed):
        if 'miss' in names[i]:
            assert all(h is  None or len(h.strip()) > 0 for h in hs)
        else:
            assert names[i] in hs

def test_multi():

    rows= [['A',None,'C'],['','b','c']]
    names = create_column_names(rows)

    h_transposed = list(map(list, zip(*rows)))
    for i, hs in enumerate(h_transposed):
        if 'miss' in names[i]:
            assert all(h is  None or len(h.strip()) > 0 for h in hs)
        else:
            assert names[i] in hs

def test_duplciates():

    rows= [['A',None,'A']]
    names = create_column_names(rows)

    h_transposed = list(map(list, zip(*rows)))
    for i, hs in enumerate(h_transposed):
        if 'miss' in names[i]:
            assert all(h is  None or len(h.strip()) > 0 for h in hs)
        else:
            assert names[i] in hs



#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from pyyacp.profiler import ColumnProfiler

from pyyacp.profiler.colum_pattern_profiler import ColumnPatternProfiler
from pyjuhelpers.timer import timer

INT='int'
FLOAT='float'
DATETIME='datetime'
DATE='date'
TIME='time'
UNICODE='unicode'

import re
types=[
    (FLOAT,re.compile('^\[?[+/-]*\]?\d+\.\d*$')),
    (INT,re.compile('^\[?[+/-]*\]?\d+$')),
    (DATETIME, re.compile('^\d{4}-\d{2}-\d{2}([C ]+\d{2}:\d{2}:\d{2})$')),
    (DATE, re.compile('^\d{4}-\d{2}-\d{2}$')), #1111-11-11
    (DATE, re.compile('^\d{4}/\d{2}/\d{2}$')), #1111/11/11
    (DATE, re.compile('^\d{2}/\d{2}/\d{4}$')),#11/11/1111

    (TIME, re.compile('^\d{2}:\d{2}(:\d{2})?$')) #1111-11-11
]

class DataTypeDetection(ColumnProfiler):

    def __init__(self):
        super(DataTypeDetection, self).__init__('dtype','data_type')

    def result_datatype(self):
        return str

    @timer(key="profile.col_dtype")
    def _profile_column(self, values, meta):

        data_type = UNICODE
        if 'pattern' not in meta:
            ColumnPatternProfiler().profile_column(values, meta)
        pattern = meta['pattern']

        gtypes = set([])
        for ptype in types:
            if pattern is not None:
                m = ptype[1].match(pattern)
                if m:
                    gtypes.add(ptype[0])
        if len(gtypes) != 1:
            data_type =  UNICODE
        if len(gtypes)==1:
            data_type = gtypes.pop()
            if data_type == INT or data_type == FLOAT:
                if 'stats_min_value' in meta:
                    smallest_first_char = meta['stats_min_value'][0]
                    if smallest_first_char == u'0':
                        # print "LEADING ZERO"
                        return UNICODE

        return data_type





import inspect

from pyjuhelpers.module_import import import_from_string
from pyyacp.config import pyyacpconfig, my_import
from pyyacp.datatable import parseDataTables
from pyyacp.profiler import Profiler, ColumnProfilerSet, ColumnProfiler, TableProfilerSet, TableProfiler
from pyyacp.profiler.profiling import get_profilers_keys, apply_profilers

from pyyacp.profiler.column_stats_profiler import ColumnStatsProfiler
from pyyacp.profiler.fdprofiler import FDProfiler
from pyyacp.profiler.cset_detection import CharacterSetProfiler
from pyyacp.profiler.colum_pattern_profiler import ColumnPatternProfiler
from pyyacp.profiler.data_type_detection import DataTypeDetection
from pyyacp.profiler.datatype_interpretation import DataTypeInterpretation
from pyyacp.profiler.distributions import CharacterDistributionProfiler
from pyyacp.profiler.distributions import BenfordsLawDistribution

from tests import create_dummy_table


def _test_profilers(table, profilers, capsys):



    ptable = apply_profilers(table, profilers=profilers)

    if isinstance(profilers, list):
        for p in profilers:
            _test_profiler(ptable   , p,capsys)
    elif isinstance(profilers, str) or isinstance(profilers, Profiler) or issubclass(profilers, Profiler):
        _test_profiler(ptable, profilers,capsys)

def _test_profiler(ptable , profiler,capsys):
    if isinstance(profiler , str):
        profiler = import_from_string(profiler)


    if inspect.isclass(profiler):
        profiler = profiler()

    if isinstance(profiler, ColumnProfilerSet) or isinstance(profiler, ColumnProfiler):

        meta = ptable.column_metadata
        for k, m in meta.items():
            for k in profiler.profiler_keys():
                assert k in m
                assert k in ptable.profiler_keys['column']
                with capsys.disabled():
                    print("{} -> {}".format(k, m[k]))


    elif isinstance(profiler, TableProfilerSet) or isinstance(profiler, TableProfiler):
        meta = ptable.table_metadata
        for k in profiler.profiler_keys():
            assert k in meta
            assert k in ptable.profiler_keys['table']
            with capsys.disabled():
                print("{} -> {}".format(k, meta[k]))


def _setup_table(tmpdir):
    d = tmpdir.mkdir("data")
    csv1 = str(d.join("table1.csv"))
    create_dummy_table(csv1)

    return csv1

def _parse_table(tmpdir,capsys):
    csv1 = _setup_table(tmpdir)


    table = parseDataTables(csv1, max_tables=1)

    with capsys.disabled():
        print(table.data.head(5))

    return table

def test_column_stats_profiler(tmpdir,capsys):

    table= _parse_table(tmpdir,capsys)

    profiler = ColumnStatsProfiler

    _test_profilers(table, profiler, capsys)

def test_FD_PROFILER(tmpdir,capsys):
    table = _parse_table(tmpdir,capsys)

    profiler = FDProfiler

    _test_profilers(table, profiler,capsys)


def test_CharacterSetProfiler(tmpdir,capsys):
    table = _parse_table(tmpdir,capsys)

    profiler = CharacterSetProfiler

    _test_profilers(table, profiler,capsys)


def test_ColumnPatternProfiler(tmpdir,capsys):
    table = _parse_table(tmpdir,capsys)

    profiler = ColumnPatternProfiler

    _test_profilers(table, profiler,capsys)


def test_DataTypeDetection(tmpdir,capsys):
    table = _parse_table(tmpdir,capsys)


    profiler = DataTypeDetection

    _test_profilers(table, profiler,capsys)


def test_DataTypeInterpretation(tmpdir, capsys):
    table = _parse_table(tmpdir, capsys)

    profiler = DataTypeInterpretation

    _test_profilers(table, profiler, capsys)


def test_CharacterDistributionProfiler(tmpdir,capsys):
    table = _parse_table(tmpdir,capsys)


    profiler = CharacterDistributionProfiler

    _test_profilers(table, profiler,capsys)


def test_BenfordsLawDistribution(tmpdir,capsys):
    table = _parse_table(tmpdir,capsys)

    profiler = BenfordsLawDistribution

    _test_profilers(table, profiler,capsys)




def test_all(tmpdir,capsys):
    table = _parse_table(tmpdir,capsys)

    profilers = [BenfordsLawDistribution

                ,CharacterDistributionProfiler
                ,DataTypeDetection
                ,ColumnPatternProfiler
                ,CharacterSetProfiler
                ,FDProfiler
                ,ColumnStatsProfiler
    ]

    _test_profilers(table, profilers,capsys)


def test_config_all(tmpdir,capsys):
    table = _parse_table(tmpdir,capsys)

    _test_profilers(table, pyyacpconfig.DEFAULT_PROFILERS,capsys)



def test_get_all_keys():
    profilers = [BenfordsLawDistribution

        , CharacterDistributionProfiler
        , DataTypeDetection
        , ColumnPatternProfiler
        , CharacterSetProfiler
        , FDProfiler
        , ColumnStatsProfiler
                 ]

    keys= get_profilers_keys(profilers)
    for p in profilers:
        p=p()
        for k in p.profiler_keys():
            if isinstance(p,TableProfiler):
                assert k in keys['table']
            elif isinstance(p,ColumnProfiler):
                assert k in keys['column']








#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals



from dtpattern.dtpattern1 import translate_all
from pyyacp.profiler import ColumnProfiler
from pyjuhelpers.timer import timer

class CharacterSetProfiler(ColumnProfiler):

    def __init__(self):
        super(CharacterSetProfiler, self).__init__('cset', 'cset')

    def result_datatype(self):
        return str

    @timer(key="profile.col_cset")
    def _profile_column(self, column, meta):

        return self._analyse_ColumnPattern(column, meta)


    def _analyse_ColumnPattern(self, column,meta):
        patterns = translate_all(column)

        res=set([])
        for p in patterns:
            s=set(p)
            res.update(s)

        return ''.join(sorted(list(set(res))))





#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import inspect

import structlog
log = structlog.get_logger()

from pyyacp.profiler.empty_cell_detection import is_not_empty


def isnamedtupleinstance(x):
    t = type(x)
    b = t.__bases__
    if len(b) != 1 or b[0] != tuple: return False
    f = getattr(t, '_fields', None)
    if not isinstance(f, tuple): return False
    return all(type(n)==str for n in f)


from abc import abstractmethod
class Profiler(object):
    def __init__(self, id, key):
        self.id=id
        self.key=key


    @abstractmethod
    def result_datatype(self):
        """

        :return: the datatype of the profiler result object
        """
        pass

    def _add_results(self, meta, results):

        if isnamedtupleinstance(results):
            for k, v in results._asdict().items():
                meta["{}_{}".format(self.key, k)] = v
        elif isinstance(results, dict):
            for k, v in results.items():
                meta["{}_{}".format(self.key, k)] = v
        else:
            meta["{}".format(self.key)] = results



    def profiler_keys(self):
        results = self.result_datatype()

        keys=[]
        if isnamedtupleinstance(results):
            for k, v in results._asdict().items():
                keys.append("{}_{}".format(self.key, k))
        elif isinstance(results, dict):
            for k, v in results.items():
                keys.append("{}_{}".format(self.key, k))
        else:
            keys.append("{}".format(self.key))

        return keys

    def isProfiled(self, meta):
        for k in self.profiler_keys():
            if k not in meta:
                return False
        return True



class TableProfiler(Profiler):
    def __init__(self, id, key):
        super(TableProfiler, self).__init__( id, key)


    def profile_table(self, table):
        if not self.isProfiled(table.table_metadata):
            result = self._profile_table(table)
            self._add_results(table.table_metadata, result)


    @abstractmethod
    def _profile_table(self, table):
        """

        :param table:
        :return: the profiler result object
        """
        pass


class ColumnProfiler(Profiler):
    def __init__(self, id, key):
        super(ColumnProfiler, self).__init__(id,key)

    def profile_column(self, column, meta):
        if not self.isProfiled(meta):
            result = self._profile_column(column, meta)
            self._add_results(meta, result)



    @abstractmethod
    def _profile_column(self, column, meta):
        """

        :param column: the column data
        :param meta: the current column metadata
        :return: the profile object
        """
        pass

class ColumnProfilerSet(object):
    """
    This class provides a wrapper for profilers which take a column as input
    """
    def __init__(self, profilers=[]):
        self.profilers=profilers


    def profile_column(self, column, meta):
        for p in self.profilers:
            if inspect.isclass(p):
                p=p()
            p.profile_column(column, meta)


class TableProfilerSet(object):
    def __init__(self, profilers=[]):
        self.profiler_factories=profilers


    def profile_table(self, table):
        for p in self.profilers:
            if inspect.isclass(p):
                p=p()
            p.profile_table(table)






import collections
import inspect

from pyjuhelpers.module_import import import_from_string
from pyjuhelpers.string_format import reindent
from pyyacp.config import pyyacpconfig
from pyyacp.datatable import DataTable
from pyyacp.profiler import Profiler, ColumnProfilerSet, ColumnProfiler, TableProfilerSet, TableProfiler

from pyjuhelpers.timer import timer
import pandas as pd

TableProfile = collections.namedtuple('TableProfile',['table_metadata','table_profile', 'col_profiles'])
class ProfiledDataTable(DataTable):

    def __init__(self, datatable):
        if datatable is not None:
            self.__dict__.update(datatable.__dict__)

        self.table_metadata = {}
        self.column_metadata = {i: {} for i in range(0, len(datatable.column_names))}
        self._profiler_keys={}

    def update_profiler_keys(self, key_dict):
        for k, v in key_dict.items():
            if k in self._profiler_keys:
                self._profiler_keys[k] += v
            else:
                self._profiler_keys[k] = v

    @property
    def profiler_keys(self):
        return self._profiler_keys

    @property
    def table_profiler_keys(self):
        return self._profiler_keys['table']

    def colum_profiles_df(self)-> pd.DataFrame:
        """

        :return: a pandas data frame
        """
        d={}
        for i in range(0, self.no_cols):
            dd=d.setdefault('col{}'.format(i+1),{})
            for a,h in enumerate(self.header_rows):
                dd['header{}'.format(a)]=h[i]
            for k,v in self.column_metadata[i].items():
                dd[k]=v

        return pd.DataFrame(d)

    @property
    def column_profiler_keys(self):
        return self._profiler_keys['column']

    def col_profile_dict(self):
        col_profiles={}
        for col, meta in self.colum_profiles_df().to_dict().items():
            col_profiles[col] = {k: meta[k] for k in self.column_profiler_keys if k in meta}
            col_profiles[col]['col'] = col
        return col_profiles

    def table_profile(self):
        return self.table_metadata

    def print_summary(self):
        super(ProfiledDataTable, self).print_summary()

        print ("#  {:=^76}".format(' TABLE META '))
        for k,v in self.table_profile().items():
            print("#  {:>15}: {}".format(k, v))

        print("#  {:=^76}".format(' COLUMN PROFILE '))

        print (reindent(self.colum_profiles_df().to_string(line_width=80), 5,prefix="#"))

        print("#  {:-^76}".format(" DATA {} ".format(self.data.shape)))

        print(reindent(self.data.head(5).to_string(index=False,line_width=80),5, prefix="#"))

    def profiledata(self):
        return TableProfile(self.metadata, self.table_profile(), self.col_profile_dict())



@timer()
def apply_profilers(table, profilers = pyyacpconfig.DEFAULT_PROFILERS ) -> ProfiledDataTable:

    if not isinstance(table, ProfiledDataTable) and isinstance(table, DataTable):
        table = ProfiledDataTable(table)

    if isinstance(profilers, list):
        for p in profilers:
            if isinstance(p, str):
                p = import_from_string(p)
            apply_profiler(table,p)

    elif isinstance(profilers, Profiler) or issubclass(profilers, Profiler):
        apply_profiler(table, profilers)

    return table

def apply_profiler(table, profiler) -> ProfiledDataTable:
    if not isinstance(table, ProfiledDataTable) and isinstance(table, DataTable):
        table = ProfiledDataTable(table)

    if inspect.isclass(profiler):
        profiler = profiler()

    table.update_profiler_keys(get_profiler_keys(profiler))

    if isinstance(profiler, ColumnProfilerSet) or isinstance(profiler, ColumnProfiler):
        for i, col in enumerate(table.columns()):
            profiler.profile_column(col, table.column_metadata[i])
    elif isinstance(profiler, TableProfilerSet) or isinstance(profiler, TableProfiler):
        profiler.profile_table(table)

    return table

def get_profilers_keys(profilers = pyyacpconfig.DEFAULT_PROFILERS ):
    keys = {}
    if isinstance(profilers, list):
        for p in profilers:
            _keys = get_profiler_keys(p)
            for k,v in _keys.items():
                if k in keys:
                    keys[k] += v
                else:
                    keys[k] = v

    elif isinstance(profilers, Profiler) or issubclass(profilers, Profiler):
        _keys = get_profiler_keys( profilers)
        for k, v in _keys.items():
            if k in keys:
                keys[k] += v
            else:
                keys[k] = v

    return keys


def get_profiler_keys(profiler):

    if isinstance(profiler, str):
        profiler = import_from_string(profiler)

    keys={}
    if inspect.isclass(profiler):
        profiler = profiler()

    if isinstance(profiler, ColumnProfilerSet) or isinstance(profiler, ColumnProfiler):
        keys['column'] = profiler.profiler_keys()

    elif isinstance(profiler, TableProfilerSet) or isinstance(profiler, TableProfiler):
        keys['table'] = profiler.profiler_keys()

    return keys





#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import hashlib
import inspect
import itertools
import logging
from io import StringIO

import anycsv
import pandas as pd
import csv

from pyjuhelpers.logging import log_func_detail
from pyjuhelpers.module_import import import_from_string
from pyyacp.config import pyyacpconfig, anycsvconfig

from pyyacp.table_structure_helper import _most_common_oneliner


from pyyacp import YACParserException

from pyjuhelpers.string_format import reindent

import structlog
log = structlog.get_logger()



def grouper(iterable, size):
    """
        batch_group generator
    :param n:
    :param iterable:
    :return:
    """
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, size))
        if not chunk:
            return
        yield chunk


@log_func_detail(log, level=logging.INFO, log_time=True)
def parseDataTables(csv,
                    skip_guess_encoding = anycsvconfig.SKIP_GUESS_ENCODING,
                    sniff_lines = anycsvconfig.NO_SNIFF_LINES,
                    max_file_size = anycsvconfig.MAX_FILE_SIZE,
                    encoding = anycsvconfig.DEFAULT_ENCODING,
                    batch_size = pyyacpconfig.BATCH_SIZE, structure_detector = pyyacpconfig.DEFAULT_STRUCTURE_DETECTOR,
                    max_tables=pyyacpconfig.MAX_TABLES, raiseError=pyyacpconfig.RAISE_ERROR):
    """

    :param csvreader:
    :param batch_size:
    :param structure_detector:
    :param max_tables:
    :param raiseError: in case we parse more than max_tables , we raise an error if flag is True
    :return:
    """

    csvreader = anycsv.reader(csv,
                           skip_guess_encoding = skip_guess_encoding,
                           sniff_lines = sniff_lines, max_file_size = max_file_size, encoding = encoding)


    if isinstance(structure_detector, str):
        structure_detector = import_from_string(structure_detector)
    if inspect.isclass(structure_detector):
        structure_detector= structure_detector()

    if isinstance(csvreader, anycsv.csv_parser.Table):
        csv=csvreader.csv
        encoding = csvreader.encoding
        dialect = csvreader.dialect
    else:
        csv = None
        encoding = None
        dialect = None


    tables = []

    cur_dt = None


    skipped = 0

    rows_to_add=[]

    row_len_groups = [] # list of row_len_groups
    rows_parsed = 0 # count to keep track how many rows we parsed already



    for g_rows in grouper(csvreader, batch_size):
        rows_parsed += len(g_rows)

        # analys the shape of the rows
        _row_lens = list(map(len, g_rows))
        _row_len_groups = [(k, sum(1 for i in g)) for k, g in itertools.groupby(_row_lens)]
        _max_len = max(_row_lens)
        _est_colNo = _most_common_oneliner(_row_lens)

        row_len_groups += _row_len_groups

        if len(_row_len_groups) == 1:
            # perfect, one table in this batch, all rows have the same length
            if cur_dt is None:
                # we have no table, this is hte first
                g_rows = list(g_rows)

                #detect comment and header based on the rows in the current batch
                comments = structure_detector.guess_description_lines( g_rows )
                header = structure_detector.guess_headers(g_rows)

                cur_dt = DataTable( csv, encoding, dialect, _est_colNo, comments=comments, headers=header)

                pos = len(comments) + len(header) #get position of data rows
                rows_to_add.extend( g_rows[pos:] ) # one group, add remaining rows as data rows

            elif _max_len == cur_dt.no_cols:
                #ok we have an active table and the current batch has the same cols as our table -> assume that they belong together
                rows_to_add.extend(g_rows)
            else:

                # not the same length, maybe different table , should not happen
                #create a new table
                # g_rows = list(g_rows)
                #
                # # detect comment and header based on the rows in the current batch
                # comments = structure_detector.guess_description_lines(g_rows)
                # header = structure_detector.guess_headers(g_rows)
                #
                # cur_dt = DataTable(csv, encoding, dialect, _est_colNo, comments=comments, headers=header)
                #
                # pos = len(comments) + len(header)  # get position of data rows
                # rows_to_add.extend(g_rows[pos:])  # one group, add remaining rows as data rows

                log.warning("NOT IMPLEMENTED", csv=csv,
                            msg="not the same length, maybe different table")
        else:
            # lets go over the groups
            # (2,30) -> belongs to old table
            # (0,1)  -> empty line
            # (1,1)  -> comment line -> flag create_new
            # (4,20) -> belongs to new table -> create new table, start parsing at (1,1)

            cur_line = 0
            create_new = False
            for i, group in enumerate(_row_len_groups):

                _cols = group[0]
                _no_rows = group[1]



                if _cols == 0:   # empty line, skip, 0 columns
                    skipped += 1
                    pass
                elif _cols == 1 and _no_rows < 3:
                    # there is a group with one element, that should be the comment lines
                    # also this means a new table
                    if i == len(_row_len_groups) - 1 and [sum(x) for x in zip(*_row_len_groups)][1] < batch_size:
                        #no idea what i do here
                        # print "SUFFIX COMMENT LINES"
                        log.warning("SUFFIX COMMENT LINES")
                    else:
                        # we have more groups to come, so lets start a new table from this line
                        if not create_new:
                            parse_start = cur_line
                        create_new = True
                else:
                    # a group with more than one column
                    start = None
                    if cur_dt is None or create_new:
                        start = cur_line
                        if create_new:
                            start = parse_start
                    elif _cols == cur_dt.no_cols:
                        #cur_dt.addRows(g_rows[cur_line:group[1]])
                        rows_to_add.extend(g_rows[cur_line:cur_line+group[1]])
                    else:
                        # seems like a new table
                        if _no_rows != 1 or (
                                        i == len(_row_len_groups) - 1 and
                                        [sum(x) for x in zip(*_row_len_groups)][1] == batch_size):
                            # more than one row
                            # OR at the end of the group and still a full batch
                            start = cur_line
                        else:
                            #print ("NOT TREATED", group[0])
                            # if only one row and (at the end of the file or in the middle of a group)
                            pass

                    if start is not None:
                        if cur_dt:
                            cur_dt.addRows(rows_to_add)
                            rows_to_add=[]

                            cur_dt.done()
                            tables.append(cur_dt)

                        _rows=g_rows[start:cur_line+_no_rows]
                        comments = structure_detector.guess_description_lines(_rows)
                        header = structure_detector.guess_headers(_rows)

                        cur_dt = DataTable( csv, encoding, dialect, _cols, comments=comments, headers=header)

                        pos = len(comments) + len(header) + start
                        end = cur_line + _no_rows

                        rows_to_add.extend(g_rows[pos:end])
                        create_new = False

                cur_line += _no_rows

    cur_dt.addRows(rows_to_add)
    rows_to_add = []
    cur_dt.done()
    tables.append(cur_dt)

    prev_group = None
    agg_groups = []
    for group in row_len_groups:
        if prev_group is not None:
            if prev_group[0] == group[0]:
                # merge
                prev_group = ( prev_group[0], prev_group[1] + group[1])
            else:
                agg_groups.append(prev_group)
                prev_group = group
        else:
            prev_group = group

    agg_groups.append(prev_group)
    log.info("TABLE SHAPE", groups=agg_groups, csv=csv)

    if len(tables) > max_tables:
        if raiseError:
            raise YACParserException("Too many tables (#{}) shapes:{}, csv:{}".format(len(tables),str(agg_groups), csv))
    log.info("Parsed table", skipped=skipped, tables=len(tables), csv=csv)

    if max_tables == 1:
        return tables[0]
    else:
        return tables


def create_column_names(header_rows, cols=None):
    if cols is None:
        cols = len(header_rows)
    names=[]
    done=False
    if len(header_rows) >= 1:
        #we found at least one header row

        #lets check if there is any header row with values in each column
        for h_row in header_rows:
            if all( h is not None and len(h.strip())>0 for h in h_row):
                #all header have a non empty value
                names = [h.strip() for h in h_row]
                done = True
                break
        if not done:
            #ok second attemp, take always the first available header
            h_transposed = list(map(list, zip(*header_rows)))
            for i, hs in enumerate(h_transposed):
                try:
                    name=next(s for s in hs if s and len(s.strip())>0)
                except:
                    name="miss{}".format(i+1)
                names.append(name)
    else:
        names = ['col{}'.format(i) for i in range(1, cols+1)]

    ##check that we do not have duplicates
    newlist = []
    for i, v in enumerate(names):
        totalcount = names.count(v)
        count = names[:i].count(v)
        newlist.append(v + str(count + 1) if totalcount > 1 else v)
    names = newlist

    return names


DataTableMetaData = collections.namedtuple('DataTableMetaData',
                                           ['csv', 'encoding', 'dialect', 'digest',
                                            'comment_rows','header_rows',
                                            'no_cols','no_rows','column_names'])


class DataTable(object):

    def __init__(self, csv, encoding, dialect, cols, comments=None, headers=None):
        self.csv=csv
        self.encoding=encoding
        self.dialect = dialect

        self.hash = hashlib.md5()
        self.digest =None

        self.comment_rows = comments if comments else []
        self.header_rows = headers if headers else []

        for c in self.comment_rows:
            self.hash.update("".join(c).encode('utf-8'))
        for c in self.header_rows:
            self.hash.update( "".join(c).encode('utf-8'))

        #store the dimensionality of the table
        self.no_cols = cols
        self.no_rows = 0

        self.column_names = create_column_names(self.header_rows,self.no_cols)

        self.data=pd.DataFrame(columns=self.column_names)

    def shape(self):
        """
        :return: tuple (rows, columns
        """
        return (self.no_rows, self.no_cols)

    @property
    def metadata(self) ->  DataTableMetaData:
        return DataTableMetaData(self.csv,
                                 self.encoding,
                                 self.dialect,
                                 self.digest,
                                 self.comment_rows,
                                 self.header_rows,
                                 self.no_cols,
                                 self.no_rows,
                                 self.column_names)

    def addRows(self, rows):
        try:
            for c in rows:
                self.hash.update("".join(c).encode('utf-8'))

            df = pd.DataFrame( list(rows), columns=self.column_names)
            self.data=pd.concat([self.data,df])
            self.no_rows = self.data.shape[0]
        except Exception as e:
            log.fatal("Exception In Adding Rows", exc_info=e)

    def rows(self):
        return [row for row in self.rowIter()]

    def rowIter(self):
        for row in self.data.itertuples():
            yield list(row[1:])


    def columnIter(self):
        for colName in self.column_names:
            yield self.data[colName].tolist()

    def columns(self):
        return [collist for collist in self.columnIter()]


    def done(self):
        self.digest = self.hash.hexdigest()
        self.hash=None
        del self.hash

    def generate(self, delimiter=',', newline='\n', header=True, comments=True):
        """
        :param delimiter: The delimiter symbol. The default is ",".
        :param newline: The line separator. The default is "\n".
        :param header: There will be a header in the output (if no header detected then col0,col1, ...)
        :param commentPrefix: The prefix for comments in the first rows. If false, any comments will be ignored.
        :return: A string representation of the CSV table
        """
        csvstream = StringIO()
        w = csv.writer(csvstream,  delimiter=str(delimiter), lineterminator=str(newline))
        c=0
        # write description lines at top
        if comments:
            for line in self.comments:
                w.writerow(line)
                c+=1
        if header:
            for header in self.header_rows:
                w.writerow([str(h) for h in header])
                c += 1
        for row in self.rowIter():
            w.writerow([str(r) for r in row])
            c += 1
        return csvstream.getvalue()


    def print_summary(self):

        print("{:#^80}".format(' TABLE STRUCTURE '))
        for k, v in self.metadata._asdict().items():
            print("#  {:>15}: {}".format(k,v))











#!/usr/bin/python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from pyyacp.profiler import ColumnProfiler
from pyyacp.profiler.colum_pattern_profiler import ColumnPatternProfiler
from pyyacp.profiler.column_stats_profiler import ColumnStatsProfiler

import re

from pyyacp.profiler.data_type_detection import DataTypeDetection

SENT=re.compile('^(C[c])+( [Cc]+)*[\.!?]?$')
ENT=re.compile('^(C[c])+( [Cc]+){2,3}$')

from pyjuhelpers.timer import timer

class DataTypeInterpretation(ColumnProfiler):

    def __init__(self):
        super(DataTypeInterpretation, self).__init__('dclass','data_class')

    def result_datatype(self):
        return str

    @timer(key="profile.col_dclass")
    def _profile_column(self, column, meta):
        if 'pattern' not in meta:
            ColumnPatternProfiler().profile_column(column, meta)
        if 'stats_max_len' not in meta:
            ColumnStatsProfiler().profile_column(column, meta)
        if 'data_type' not in meta:
            DataTypeDetection().profile_column(column, meta)

        return self._analyse_ColumnPattern(column,meta)

    def _analyse_ColumnPattern(self, column,meta):
        from pyyacp.profiler.data_type_detection import DATETIME, INT, UNICODE
        patterns = meta['pattern']
        data_type=meta['data_type']
        #print meta
        min, mean, max = meta['stats_max_len'], meta['stats_mean_len'], meta['stats_min_len']
        fixed_len = min ==max
        unique = meta['stats_distinct'] + meta['stats_empty'] == meta['stats_num_rows']

        if data_type==DATETIME:
            return data_type.upper()

        if meta['stats_empty'] == meta['stats_num_rows']:
            return "EMPTY"

        s=''
        if data_type == INT:
            s+='NUM'
        elif data_type == UNICODE:
            if patterns:
                if SENT.match(patterns):
                    s+='SENTENCE'
                elif ENT.match(patterns):
                    s += 'ENTITY'
                else:
                    s += 'UNI'
            else:
                s += 'UNI'
        else:
            s = 'UNDEF'

        if fixed_len:
            s+='_ID'
        else:
            s+='_VAR'

        if unique:
            s+='_UNIQ'

        if meta['stats_distinct'] == 1:
            s+="_SING"
        elif (meta['stats_num_rows'] > 10 and (meta['stats_distinct']/float(meta['stats_num_rows']))<0.1) \
                or (meta['stats_num_rows'] < 10 and meta['stats_distinct']<=2):
            s+='_CAT'
        return s



import pytest
from pyyacp.datatable import parseDataTables


t1=[['A','B','C'],
    ['1','2','3'],
    ['2','3','4']
    ]
t2=[
    ['A', 'B'],
    ['1', '2'],
    ['2', '3'],
]


testdata=[ t1,t2]
@pytest.mark.parametrize("table", testdata)
def test_simple(table):

    csvreader=table

    tables = parseDataTables(csvreader)
    assert len(tables) ==1
    table = tables[0]
    assert table.no_cols == len(csvreader[0])
    assert table.no_rows == len(csvreader)-1
    assert len(table.header_rows) == 1
    assert len(table.comment_rows) == 0

t1=[['6','5','1'],
    ['1','2','3'],
    ['2','3','4']
    ]
t2=[
    ['2', '4'],
    ['1', '2'],
    ['2', '3'],
]


testdata=[ t1,t2]
@pytest.mark.parametrize("table", testdata)
def test_simple_noheader(table):

    csvreader=table
    tables = parseDataTables(csvreader)
    assert len(tables) ==1
    table = tables[0]
    assert table.no_cols == len(csvreader[0])
    assert table.no_rows == len(csvreader)
    assert len(table.header_rows) == 0
    assert len(table.comment_rows) == 0


t1=[
    ['Comment'],
    ['6','5','1'],
    ['1','2','3'],
    ['2','3','4']
    ]
t2=[
    ['Comment'],
    ['2', '4'],
    ['1', '2'],
    ['2', '3'],
]


testdata=[ t1,t2]
@pytest.mark.parametrize("table", testdata)
def test_simple_comment_no_header(table):

    csvreader=table
    tables = parseDataTables(csvreader)
    assert len(tables) == 1
    table = tables[0]
    assert table.no_cols == len(csvreader[1])
    assert table.no_rows == len(csvreader)-1
    assert len(table.header_rows) == 0
    assert len(table.comment_rows) == 1

def test_two_tables():
    csvreader = [
        ['A', 'B', 'C'],
        ['1', '2', '3'],
        ['2', '3', '4'],
        ['A', 'B'],
        ['1', '2'],
        ['2', '3'],

    ]
    tables = parseDataTables(csvreader)
    assert len(tables) == 2
    table = tables[0]
    assert table.no_cols == 3
    assert table.no_rows == 2
    assert len(table.header_rows) == 1


def test_two_tables_with_comments_empty_line():
    csvreader = [
        ['This is a comment'],
        ['2', '3', '4'],
        ['1', '2', '3'],
        ['2', '3', '4'],
        [],
        ['A', 'B'],
        ['1', '2'],
        ['2', '3'],

    ]
    tables = parseDataTables(csvreader)
    assert len(tables) == 2

    table = tables[0]
    assert table.no_cols == 3
    assert table.no_rows == 3
    assert len(table.header_rows) == 0
    assert len(table.comment_rows) == 1


    table = tables[1]
    assert table.no_cols == 2
    assert table.no_rows == 2
    assert len(table.header_rows) == 1
    assert len(table.comment_rows) == 0

